# NOTE: some functions in random module are not suitable to cryptography
# to be cryptography safe, use random.SystemRandom that relies on /dev/urandom
import random
import os

def generate_random_MAC(DEBUG=False):
    """ generate a random MAC address representation as a string, with a valid vendor OUI
    A MAC number is 6 groups of hexadecimal digits ie 6 groups of octets ie 48 bits
    MAC numbers are usually displayed as oct1:oct2:oct3:oct4:oct4:oct6
    The first 3 octets are Organizationally Unique Identifier [OUI]
    To generate a plausible random MAC number, the OUI is extracted from an OUI list
    The OUI list used by default is the one of macchanger (see https://github.com/alobbs/macchanger/tree/master/data )
    """

    # random number generator safe for cryptography --------------------------------
    cryptogen = random.SystemRandom()

    # valid OUI --------------------------------------------------------------------
    # use OUI.list from macchanger

    # path to the data relative to the position of this file
    # file_valid_MAC = os.path.abspath(os.path.join(os.path.dirname(__file__), '/../data/OUI.list'))
    path_to_this_file = os.path.dirname(__file__)
    file_valid_MAC = path_to_this_file + '/../data/OUI.list'

    if DEBUG:
        print file_valid_MAC

    # find number of lines
    number_of_lines = sum(1 for line in open(file_valid_MAC))
    if DEBUG:
        print number_of_lines

    # select one line randomly
    number_line_to_use = cryptogen.randrange(number_of_lines)
    if DEBUG:
        print number_line_to_use

    # extract the corresponding line
    with open(file_valid_MAC) as fp:
        for i, line in enumerate(fp):
            if i == (number_line_to_use-1):
                line_to_use = line
                break

    print "Using OUI: " + line_to_use[:-1]

    # extract the corresponding MAC address prefixe
    MAC_address_prefix = line_to_use[0:2]+":"+line_to_use[3:5]+":"+line_to_use[6:8]
    if DEBUG:
        print MAC_address_prefix

    # complete address -------------------------------------------------------------
    complete_random_realistic_MAC_address = MAC_address_prefix+":%02x:%02x:%02x" % (
            cryptogen.randrange(0, 255),
            cryptogen.randrange(0, 255),
            cryptogen.randrange(0, 255),
            )

    if DEBUG:
        print complete_random_realistic_MAC_address

    complete_random_realistic_MAC_address_uppercase = complete_random_realistic_MAC_address.upper()

    if DEBUG:
        print complete_random_realistic_MAC_address_uppercase

    return(complete_random_realistic_MAC_address_uppercase)



import subprocess

def subprocess_cmd(command,DEBUG=False):
    """execute a bash command and return output"""
    if DEBUG:
        print "command: " + command
    process = subprocess.Popen(command,stdout=subprocess.PIPE, shell=True)
    proc_stdout = process.communicate()[0].strip()
    if DEBUG:
        print "out: " + proc_stdout
    return(proc_stdout)



# extract information from nmcli through bash calls
from execute_bash_command import *

# list of commands -------------------------------------------------------------

# command line to print the name of the saved connections:
# print connections     | first work is name | not first line (label)
# nmcli connection show | awk '{print $1}'   | tail -n +2
bash_list_saved_connections = "nmcli connection show | awk '{print $1}'   | tail -n +2"

# command line to print the name of the connection types
# print connections     | third work is type | not first line (label)
# nmcli connection show | awk '{print $3}'   | tail -n +2
bash_list_saved_connection_types = "nmcli connection show | awk '{print $3}'   | tail -n +2"

# command line to print the list of physical devices
# print status devices  | first is device name | nor first line (label)
bash_list_physical_devices = "nmcli dev status | awk '{print $1}'   | tail -n +2"

# command line to print the list of connection types
# print status devices  | second is type | nor first line (label)
bash_list_physical_devices_type = "nmcli dev status | awk '{print $2}'   | tail -n +2"

# print the MAC address through nmcli
# nmcli connection show Get-1b7f62 | grep mac | head -n 2

def parse_output_bash(network_names,DEBUG=False):
    """turn the string of bash network outputs into a list of python name strings"""
    list_network_words = []
    current_word = []

    if DEBUG:
        print "Into parser: "
        print network_names

    for current_char in network_names:
        if current_char == '\n':
            list_network_words.append(''.join(current_word))
            current_word = []
        else:
            current_word.append(current_char)

    # do not forget the last one!
    list_network_words.append(''.join(current_word))

    if DEBUG:
        print "Out of parser: "
        print list_network_words

    return list_network_words

def index_all_occurences(list_values,occurence):
    """returns the list of indexes at which occurence appears in list_values
    work like .index() except that return a list of all occurences"""

    list_occurences_index = [i for i, j in enumerate(list_values) if j == occurence]
    return list_occurences_index

def nmcli_show_connections():
    """a wrapper around nmcli connection show"""

    command_output = subprocess_cmd("nmcli connection show")
    print command_output
    return command_output

def show_MAC_address_connection(connection, DEBUG=False):
    """return information about MAC and cloned MAC address"""

    command = "nmcli connection show " + connection + " | grep mac | head -n 2"
    command_output = subprocess_cmd(command)

    if DEBUG:
        print "MAC address information about connection: " + connection
        print command_output

    return(command_output)

def nmcli_saved_connections(DEBUG=False):
    """returns the name of the saved connections as a Python list"""

    command_output = subprocess_cmd(bash_list_saved_connections)
    result = parse_output_bash(command_output,DEBUG=DEBUG)

    return result

def nmcli_types(DEBUG=False):
    """return the connection types as a Python list, corresponding to each connection saved"""

    command_output = subprocess_cmd(bash_list_saved_connection_types)
    result = parse_output_bash(command_output,DEBUG=DEBUG)

    return result

def nmcli_device_wifi(DEBUG=False):
    """returns name of physical devices giving wifi access"""

    command_list_devices = subprocess_cmd(bash_list_physical_devices)
    list_devices = parse_output_bash(command_list_devices,DEBUG=DEBUG)

    command_list_devices_types = subprocess_cmd(bash_list_physical_devices_type)
    list_devices_type = parse_output_bash(command_list_devices_types,DEBUG=DEBUG)

    # find the one that have type wifi
    list_index_wifi = index_all_occurences(list_devices_type,"wifi")

    # return the corresponding devices
    return [list_devices[i] for i in list_index_wifi]

def nmcli_available_wifi(print_out=True):
    """print details about available Wifi networks"""

    output = subprocess_cmd("nmcli dev wifi list")

    if print_out:
        print output



# files with help

# help about nmcli
help_nmcli = """
A few helpful nmcli commands reminder:

    - nmcli dev status
    view the different interfaces and active connections

    - nmcli dev wifi list
    view the available WiFi points

    - nmcli radio wifi [on | off ]
    change the WiFi radio state
"""



#some command line interface for randommac

import argparse
from randommac_interface import *

def main():
    """the function to be called from terminal"""

    parser = argparse.ArgumentParser()

    # optional argument for launching interactive interface ------------------------
    parser.add_argument("-i","--interactive", help="launch the interactive randommac",
                        action="store_true")

    # optional argument for outputting nmcli information ---------------------------
    parser.add_argument("-n","--nmcliinfo", help="output information from nmcli",
                        action="store_true")

    # optional argument for showing information about one saved network ------------
    parser.add_argument("-d","--details", type=str, help="show details of one saved network, DETAILS is network NAME for which to show details, from nmcli")

    # optional argument to ask for a random MAC address ----------------------------
    parser.add_argument("-r","--randomaddress", help="output a random MAC address",
                        action="store_true")

    # optional argument to spoof a network -----------------------------------------
    parser.add_argument("-s","--spoof", type=str, help="spoof a saved network, SPOOF is network NAME to spoof from nmcli")

    # optional argument to connect and spoof to a new network ----------------------
    parser.add_argument("-c","--connectspoof", type=str, help="connect and spoof to a wifi network, CONNECTSPOOF is wifi network NAME to connect and spoof from nmcli")
    # additional optional argument for giving also a password
    parser.add_argument("-p","--password", type=str, help="password to connect and spoof to a wifi network, PASSWORD is the password to use on the network described by --connectspoof")


    args = parser.parse_args()

    if args.interactive:
       print "Launch interactive interface"
       interactive_interface()

    elif args.nmcliinfo:
        print "Output nmcli information"
        print "---------------------------------------------------------------------"
        print "Summary saved networks:"
        nmcli_show_connections()
        print "---------------------------------------------------------------------"
        print "All available Wifi networks:"
        nmcli_available_wifi()

    elif args.details:
        out = show_MAC_address_connection(args.details,DEBUG=True)

    elif args.randomaddress:
        address = generate_random_MAC(DEBUG=False)
        print address

    elif args.spoof:
        out = change_MAC_network([args.spoof])

    elif args.connectspoof:
        if args.password:
            # connect with a password
            add_new_wifi_spoofed_connection(args.connectspoof,connection_password=args.password)
        else:
            # connect without password
            add_new_wifi_spoofed_connection(args.connectspoof)

    else:
        print "Randommac, python wrappers on nmcli for easy MAC spoofing. Try -h for help."



# do the command line interaction for setting a new MAC with the network manager through nmcli
from execute_bash_command import *
from extract_information_from_nmcli import *
from generate_random_MAC import *


class ExceptionNoWifi(Exception):
    pass

# structure of the command to enter / save a cloned mac address using nmcli ----
# example:   nmcli connection modify Get-1b7f62 802-11-wireless.cloned-mac-address 11:22:33:44:55:66
# structure: nmcli connection modify name       type           .cloned-mac-address cloned_addresss

################################################################################
# modify the MAC address of an already existing network
def spoof_MAC_address(network_name,network_type,MAC_address, PRINT=False,DEBUG=False):
    """spoof MAC address corresponding to network name on interface type"""

    if PRINT:
        print "*********************************************************************"
        print "spoofing on network name: " + network_name
        print "MAC addresses before: -----------------------------------------------"
        show_MAC_address_connection(network_name,True)
        print "---------------------------------------------------------------------"

    # command to issue
    command_spoof = "nmcli connection modify " + network_name + " " + network_type + ".cloned-mac-address " + MAC_address

    # execute
    command_output = subprocess_cmd(command_spoof,DEBUG)

    if PRINT:
        print "MAC addresses after: ------------------------------------------------"
        show_MAC_address_connection(network_name,True)
        print "---------------------------------------------------------------------"
        print "*********************************************************************"

    return command_output

################################################################################




# provide a simple interface to randommac, both as python functions and command line interface
from generate_random_MAC import *
from spoof_MAC_address import *
from extract_information_from_nmcli import *
from execute_bash_command import *
import time
import random

################################################################################
# NOTE:
################################################################################

# ------------------------------------------------------------------------------
# PYTHON INTERFACE -------------------------------------------------------------
# ------------------------------------------------------------------------------

# display a brand new random spoofed MAC address
def display_random_MAC():
    """display and return a random MAC address"""
    random_MAC = generate_random_MAC()
    print random_MAC
    return random_MAC

# add a new network with random cloned mac address since beginning -------------
def add_new_wifi_spoofed_connection(connection_name,connection_password=""):
    """ add a new wifi connection that is spoofed since the beginning"""

    list_wifi_ifnames = nmcli_device_wifi()
    if len(list_wifi_ifnames) == 0:
        print "Found no Wifi device, abort"
        raise ExceptionNoWifi()
    elif len(list_wifi_ifnames) > 1:
        print "Found more than one Wifi device: "
        print list_wifi_ifnames
        ifname_connection = raw_input("Enter name of device to use: ")
    else:
        ifname_connection = list_wifi_ifnames[0]

    # generate a random MAC address
    random_mac = generate_random_MAC()

    # add the connection
    command_add_connection = "nmcli con add con-name " + connection_name + " ifname " + ifname_connection + " type wifi ssid " + connection_name + " cloned-mac " + random_mac

    print "Command create new network: "
    print command_add_connection
    output = subprocess_cmd(command_add_connection)
    print output

    if connection_password:
        # add password
        print "Add password type WPA2"
        command = "nmcli con modify " + connection_name + " wifi-sec.key-mgmt wpa-psk"
        output = subprocess_cmd(command)
        print command
        command = "nmcli con modify " + connection_name + " wifi-sec.psk " + connection_password
        output = subprocess_cmd(command)
        print command

# change MAC address randomly, on all saved networks ---------------------------
def change_MAC_all_saved_networks(except_networks_list=[],PRINT=True):
    """change for random MAC address for all saved networks except those in except
    Each network gets a different MAC address"""

    # deactivate all connections
    subprocess_cmd("nmcli networking off")

    # display / generate informatio from nmcli
    network_names = nmcli_saved_connections()
    network_types = nmcli_types()

    # set / change the cloned MAC address on all the connections
    for connection_number, network_name in enumerate(network_names):
        if network_name in except_networks_list:
            pass
        else:
            network_type = network_types[connection_number]
            random_MAC = generate_random_MAC()
            spoof_MAC_address(network_name,network_type,random_MAC,PRINT=PRINT)

    # reactivate all connections
    subprocess_cmd("nmcli networking on")

# change MAC address randomly, on all saved networks, same MAC on each ---------
# physical interface
def change_MAC_all_saved_networks_uniform_over_device(except_networks_list=[],PRINT=True):
    """change for random MAC for all saved networks except those in except
    All networks on the same device get the same MAC address"""

    # deactivate all connections
    subprocess_cmd("nmcli networking off")

    # display / generate informatio from nmcli
    network_names = nmcli_saved_connections()
    network_types = nmcli_types()

    # list of devices already met
    list_devices_met = []
    # list of MAC addresses corresponding to each device
    list_devices_met_addresses = []

    # set / change the cloned MAC address on all the connections
    for connection_number, network_name in enumerate(network_names):
        network_type = network_types[connection_number]

        if network_name in except_networks_list:
            pass

        else:
            if not network_type in list_devices_met_addresses:
                # create a new address for this type of network
                random_MAC = generate_random_MAC()
                list_devices_met.append(network_type)
                list_devices_met_addresses.append(random_MAC)

            # put the right cloned MAC address
            current_MAC_index = list_devices_met.index(network_type)
            current_MAC = list_devices_met_addresses[current_MAC_index]
            spoof_MAC_address(network_name,network_type,current_MAC,PRINT=PRINT)

    # deactivate all connections
    subprocess_cmd("nmcli networking on")

# change MAC address randomly, on a list of saved networks ---------------------
def change_MAC_network(network_name_list,PRINT=True):
    """change the MAC address for the network name list
    each network name gets a different MAC address"""

    # deactivate all connections
    subprocess_cmd("nmcli networking off")

    # display / generate informatio from nmcli
    network_names = nmcli_saved_connections()
    network_types = nmcli_types()

    # set / change the cloned MAC address on all the connections in network_name_list
    for connection_number, network_name in enumerate(network_names):
        if network_name in network_name_list:
            network_type = network_types[connection_number]
            random_MAC = generate_random_MAC()
            spoof_MAC_address(network_name,network_type,random_MAC,PRINT=PRINT)

    # deactivate all connections
    subprocess_cmd("nmcli networking on")

# change MAC address randomly, on all saved network for one physical -----------
# interface
def change_MAC_interface(interface_name_list,PRINT=True):
    """change the MAC address for the interface name list
    each network gets a different MAC addresss"""

    # deactivate all connections
    subprocess_cmd("nmcli networking off")

    # display / generate informatio from nmcli
    network_names = nmcli_saved_connections()
    network_types = nmcli_types()

    # set / change the cloned MAC address on all the connections in network_name_list
    for connection_number, network_type in enumerate(network_types):
        if network_type in interface_name_list:
            network_name = network_names[connection_number]
            random_MAC = generate_random_MAC()
            spoof_MAC_address(network_name,network_type,random_MAC,PRINT=PRINT)

    # deactivate all connections
    subprocess_cmd("nmcli networking off")

# periodically perform one of the random MAC changes functions -----------------
def periodically_perform_command(time_step,command):
    """periodically perform command, waiting (sleep) time_step between calls
    this is Ok as long as drift is ok in update"""

    eval(command)
    time.sleep(time_step)

# perform at random time one of the random MAC changes functions ---------------
def randomly_perform_command(min_time,max_time,command):
    """performs at random time, uniform in [min_time,max_time], the command"""

    eval(command)
    time.sleep(random.uniform(min_time,max_time))

# a step by step interface to help users
def interactive_interface():
    """a step by step interface to help the user"""

    string_welcome = """\
###########################################################
########## Welcome to step-by-step PyMacChanger ###########
###########################################################
    """
    print string_welcome

    print "---------------------------------------------------------------------"
    print "Summary saved networks:"
    nmcli_show_connections()

    print "---------------------------------------------------------------------"
    print "All available Wifi networks:"
    nmcli_available_wifi()

    print "---------------------------------------------------------------------"
    string_summary_functions = """\
    Summary of some of the available functionalities:
- [0] create spoofed connection on a new WiFi network
- [1] spoof existing connection
- [2] spoof all existing connections on an interface
- [3] start periodic spoofing
- [4] start random time spoofing
- [q] exit interactive interface
    """
    print string_summary_functions
    print "---------------------------------------------------------------------"

    continue_interactive = True

    while continue_interactive:

        function = raw_input("Choose functionality [0-4] or [q]: ")

        string_summary_spoofed_new_wifi = """\
    Create a spoofed connection on a new Wifi network
The aim is to never display your true MAC
See example: pyton_new_spoofed_connection.py
Main function to use: add_new_wifi_spoofed_connection()
Command line interface example:
    """

        string_summary_spoof_existing = """\
    Spoof an existing connection on any interface
The aim is to spoof (or modify spoofing) of an existing connection
See example: python_spoof_existing_connection.py
Main function to use: change_MAC_network()
Command line interface example:
        """

        if function == 'q':
            continue_interactive = False

        elif function == '0':
            print "---------------------------------------------------------------------"
            print string_summary_spoofed_new_wifi
            test = raw_input("Test function [y/n]: ")
            if test == 'y':
                network = raw_input("Enter new network name to spoof: ")
                password = raw_input("Enter WPA / WPA2 password (empty string for none): ")
                add_new_wifi_spoofed_connection(network,password)
            print "---------------------------------------------------------------------"

        elif function == '1':
            print "---------------------------------------------------------------------"
            print string_summary_spoof_existing
            test = raw_input("Test function [y/n]: ")
            if test == 'y':
                name_connection = raw_input("Enter name of existing connection to spoof: ")
                out = change_MAC_network([name_connection])
            print "---------------------------------------------------------------------"
